# Note that much of the time, something that's a singleton (eg: a string) can also be an iterable. Where this isn't the case, it probably should be.
# In your search field names, don't start with an underscore ('_') as that's reserved.
# We STRONGLY recommend explicitly declaring your search field names, as it makes the resultant search system more useful to users. In some cases you don't need it, but that's rare.
# Note that strictly you can have a callable as a django_field directly. In this case, it will be called with a parameter of None to generate the search field name (well, part of it - it only needs to be unique to the class). But don't do this, it's ugly.
# When auto-generating, we use '.' to separate bits of things where possible, and '__' where we require \w only.

# FIXME: document the query() method added to managers
# FIXME: make query() do pagination properly, on top of anything Flax chooses to offer us (currently Flax gives us nothing)
# FIXME: detect and resolve circular cascades

# TODO: make it possible to index an individual model to more than one database. (Probably multiple explicit indexers.)
# TODO: reverse cascades, so you can put searchable stuff into your Profile model, but have it index stuff from the User. (Also just easier in general, although I can't see how to make it as powerful as normal cascades.)
# TODO: if you change the django_fields, searchify should "want" to reindex, with suitable options to tell it not to; perhaps a hash of the config (and allow it to be set explicitly for people who want to manage this themselves)

from index import register_indexer, autodiscover, reindex, Indexer, get_searcher



"""Initialisation for the searchify app.

Despite the name, this file doesn't actually contain any models. It's just used
to contain initialisation for the searchify app, which is called at Django
setup time.

"""

from django.conf import settings
from hooks import connect_signals
from index import autodiscover

if hasattr(settings, 'ENABLE_SEARCHIFY') and settings.ENABLE_SEARCHIFY:
    connect_signals()
    autodiscover()



"""
Searchify works by having a suitable subclass of Searchable available for each model we care about.

This can be set up by one of:

 - Place a nested class in the model class called `Indexer`. (A single instance
   will be created and registered for the model.)
 - Call `register_indexer` directly with an instance.

"""

import copy
import sys
import time

from django.db import models
from django.conf import settings

import search # for make_searcher
from clients import Client
from utils import get_indexer, get_searcher, get_typename_from_object

client = Client()

class SearchifyOptions(object):
    def __init__(self, indexer=None):
        self.indexer = indexer

# Map search index -> list of models which have an indexer for that index.
_index_models = {} 

def register_indexer(model, indexer):
    """Register an indexer on a model.

    This stores the indexer in a property on the model.
    """
    
    if not hasattr(settings, 'ENABLE_SEARCHIFY') or not settings.ENABLE_SEARCHIFY:
        return

    if not hasattr(model, '_searchify'):
        model._searchify = SearchifyOptions()
    model._searchify.indexer = indexer
    indexer.model = model
    if indexer.index:
        _index_models.setdefault(indexer.index, []).append(model)

        # indexer.managers is a list of attribute names (eg: ['objects']) for managers we want to
        # decorate
        for manager in indexer.managers:
            manager = getattr(model, manager)
            manager.query = indexer.make_searcher(manager)

_ensure_dbs_exist = True
def autodiscover(verbose=None, ensure_dbs_exist=None):
    """Automatically register all models with Indexer subclasses.

    Romp through the models, creating instances of their Indexer subclasses as
    required.

    This must be called, even if you've registered the indexers yourself, since
    it sets up the search databases.
    """

    # if we get called once with ensure_dbs_exist=False, then we don't want
    # to ensure them later. This is an incredible hack to avoid having to put
    # autodiscovery into urls.py and still make it practical to have a simple
    # reindex command.
    #
    # Note we do it here so that commands that want to can defer model validation;
    # if we leave it until after the call to models.get_models() then model
    # validation will have happened and we'll have been called from lib.searchify.models.
    global _ensure_dbs_exist
    if ensure_dbs_exist is None:
        ensure_dbs_exist = _ensure_dbs_exist
    _ensure_dbs_exist = ensure_dbs_exist

    if not hasattr(settings, 'ENABLE_SEARCHIFY') or not settings.ENABLE_SEARCHIFY:
        return

    for model in models.get_models():
        if not hasattr(model, '_searchify') and hasattr(model, 'Indexer'):
            # auto-register
            if verbose:
                verbose.write("Auto-creating indexer instance for class %s\n" %
                              model)
            register_indexer(model, model.Indexer(model))

    if not ensure_dbs_exist:
        return

    # Now loop through and ensure each model has a mapping.
    for index, modellist in _index_models.items():
        for model in modellist:
            indexer = get_indexer(model)
            if indexer.get_current_mapping() is None:
                print >>sys.stderr, ("Mapping not stored for %r - need to run "
                            "reindex command" % indexer.get_typename(model))
                del _index_models[index]
                break

def reindex(indices):
    """Reindex the named indices, or all indices if none are named.

    The index is rebuilt from scratch with a new suffix, and the alias is then
    changed to point to the new index, so existing searchers should not be
    disrupted.
    """
    
    if not hasattr(settings, 'ENABLE_SEARCHIFY') or not settings.ENABLE_SEARCHIFY:
        return
    
    suffix = '_' + hex(int(time.time()))[2:]
    if not indices:
        indices = _index_models.keys()
    for indexname in indices:
        reindex_index(indexname, suffix)

def reindex_index(indexname, suffix):
    """Reindex a named index.
    """
    
    if not hasattr(settings, 'ENABLE_SEARCHIFY') or not settings.ENABLE_SEARCHIFY:
        return
    
    models = _index_models.get(indexname, None)
    if models is None:
        raise KeyError("Index %r is not known" % indexname)
    try:

        # Get the index-wide settings.
        index_settings = {}
        def merge_dicts(path, a, b):
            for (k, v) in b.iteritems():
                if k not in a:
                    a[k] = v
                    continue
                if isinstance(v, dict):
                    merge_dicts('%s.%s' % (path, k), a[k], v)
                    continue
                if a[k] == v:
                    continue
                raise ValueError("Conflicting values in index_settings (at %s)" % path[1:])
        for model in models:
            indexer = get_indexer(model)
            merge_dicts('.', index_settings, indexer.index_settings)

        created = False
        for model in models:
            print "Indexing %s to %s, using suffix %s" % (model, indexname, suffix)
            indexer = get_indexer(model)
            try:
                indexer.client.set_suffix(suffix)
                if not created:
                    #print "Creating index with settings %r" % index_settings
                    indexer.client.create_index(index_settings)
                    created = True
                indexer.apply_mapping()
                indexer.index_all(with_cascade=False)
            finally:
                indexer.client.set_suffix()
            indexer.client.flush()

        # Get the old value of the alias.
        try:
            old_index = client.get_alias(indexname)[0]
        except IndexError:
            old_index = None
        if old_index == indexname:
            # Old index wasn't an alias; we have to delete it and then set the
            # new alias for it.
            print "Warning: no alias in use, so must delete in-use index"
            old_index = None
            client.delete_index(indexname)
        print "Setting alias to make new index live"
        client.set_alias(indexname, indexname + suffix)
    except:
        try:
            client.delete_index(indexname + suffix)
        except Exception:
            # Ignore any normal exceptions, so we report the original error.
            pass
        raise
    if old_index:
        print "Removing old index: %s" % old_index
        client.delete_index(old_index)

class Indexer(object):
    """Main indexer superclass, controlling search indexing for a model.

    Most of the indexing behaviour is in here (except for the index clients).

    Typically you won't have to do much here, just subclass and set index,
    fields and perhaps cascades. However you can override any part of the
    indexing process if needed.

    """

    # index is the name of the index that this model should be indexed to.
    # The default is None, meaning that the model will not be indexed.
    index = None

    # Fields is a list of fields to be indexed.
    index_settings = {} # A dictionary of engine specific index-level settings.
    fields = []
    cascades = [] # no cascades
    managers = [] # don't create searcher by default (still pondering details, and searchers unported to class approach)
    defaults = {}

    def __init__(self, model):
        self.model = model
        if self.index:
            self.client = client.indexer(self.index)

    def reindex_on_cascade(self, cascade_from, cascade_to):
        """
        Should we reindex cascade_to when we've just reindexed cascade_from?
        Called on the indexer for cascade_to.
        """

        return True

    def should_be_in_index(self, instance):
        """
        Should we be in the index, based on the instance's current state?

        (For instance, this might want to return `not self.deleted`.)
        """

        return True

    def get_searcher(self):
        return self.client.get_searcher()

    def index_all(self, with_cascade=True):
        """Index or reindex all the instances of this model.

        If with_cascade is True, the cascade of instances depending on this
        instance will also be traversed, to update any search data built from
        these instances.

        """
        from django.db import connection
        for inst in self.model.objects.all():
            self.index_instance(inst, with_cascade)
            del inst
            connection.queries = []
        self.client.flush()

    def index_instance(self, instance, with_cascade=True):
        """Index or reindex an instance.

        If with_cascade is True, the cascade of instances depending on this
        instance will also be traversed, to update any search data built from
        these instances.

        """
        if self.index:
            if not self.should_be_in_index(instance):
                self.client.delete(self.get_typename(instance),
                                   self.get_docid(instance))
            else:
                dret = self.get_index_data(instance)
                if dret is not None:
                    (doc_type, docid, fielddata) = dret
                    self.client.add(fielddata, doc_type=doc_type, docid=docid)
        if with_cascade:
            self.cascade(instance)
        if self.index:
            self.client.flush()

    def cascade(self, instance):
        """Cascade the index from this instance to others that depend on it.

        This causes index_instance() to be called on each instance that depends
        on the instance supplied.

        """
        for descriptor in self.cascades:
            cascade_inst = None
            # find the instance we're being told to cascade the reindex onto
            try:
                if callable(descriptor):
                    cascade_inst = descriptor(instance)
                elif isinstance(descriptor, str):
                    cascade_inst = getattr(instance, descriptor)
            except:
                cascade_inst = None
            # if we found one, check if it's searchable, check if it
            # wants to accept the cascade, and if so, reindex it
            if cascade_inst:
                # If it's not an iterable already, make it into one
                if not hasattr(cascade_inst, '__iter__'):
                    cascade_insts = [cascade_inst]
                else:
                    cascade_insts = cascade_inst
                for cascade_inst in cascade_insts:
                    indexer = get_indexer(cascade_inst)
                    if indexer and indexer.reindex_on_cascade(instance, cascade_inst):
                        indexer.index_instance(cascade_inst, with_cascade=False)

    def delete(self, instance):
        """Delete an instance from the (relevant) search index.

        """

        if self.index:
            self.client.delete(self.get_typename(instance),
                               self.get_docid(instance))
            self.client.flush()

    def get_typename(self, instance):
        """Generate a type name for use in the search database.

        Default is in the format '<app_label>.<object_name>'

        """
        return get_typename_from_object(instance)

    def get_docid(self, instance):
        """Generate a docid for use as a search database identifier.

        Default is in the format ('<app_label>.<object_name>', '<pk>').

        """
        return '%s' % (instance.pk, )

    def get_index_data(self, instance):
        """Get the data to be indexed for an instance.

        Given a Django model instance, return a unique identifier and a
        dictionary of search fields mapping to lists of data, or None.

        """
        if not self.fields:
            return None

        outfields = {}

        for field in self.fields:
            (django_field_list, index_fieldname, index_config) = self.get_details(field)
            # print "indexing %s (%s)" % (instance, index_fieldname,)
            interim_data = map(lambda x: self.get_field_input(instance, x), django_field_list)
            # print '>>>' + str(interim_data)
            outfields[index_fieldname] = reduce(lambda x,y: list(x) + list(y), interim_data)

        return (self.get_typename(instance), self.get_docid(instance),
                outfields)


    def get_details(self, field):
        """
        Return (django_field_list, search field name, config) for a particular search field (which is a string, dict or possibly callable).

        Performs auto-generation of search field names as needed (generally it's not advised not to do this with a plain callable; use a
        dictionary, which makes things more clear).

        If the field is a dictionary, it will have:
            django_fields   list of django fields / callables
            field_name      single search field to index into
            config          additional config options pass through to indexing client on init

        If not, then django_fields is a list of it, and field_name is generated from:
            field is str    letters of field (eg: my_field -> myfield)
            field is callable
                            field(None)

        (Note that all these fields are in self.fields.)
        """

        if type(field) is dict:
            django_field_list = field['django_fields']
            index_fieldname = field.get('field_name')
        else:
            django_field_list = [field]
            index_fieldname = None

        if index_fieldname == None:
            if isinstance(django_field_list[0], str):
                field_specific_name = django_field_list[0]
            elif callable(django_field_list[0]):
                field_specific_name = django_field_list[0](None)
            index_fieldname = filter(lambda x: x.isalpha(), field_specific_name)

        if type(field) is dict:
            return (django_field_list, index_fieldname, field.get('config', {}))
        else:
            return (django_field_list, index_fieldname, {})


    def get_field_input(self, instance, django_field):
        """
        Given a single Django field descriptor (string or callable), generate a list of data to input to the search field.

        Converters allow Django ORM types to be modified automatically (eg: returning DateTimeField in a useful format).
        Currently, converters are embedded here, which isn't helpful.
        """

        # must return an iterable; django_field is str (name) or callable
        if isinstance(django_field, str):
            #print 'trying as str'
            #print '.name = %s' % instance.name
            #print 'getattr(,"name") = %s' % getattr(instance, 'name')
            val = getattr(instance, django_field)

            def datetime_converter(d):
                return unicode(d.date())

            # FIXME: converters should be on the class, not embedded in this method
            converters = { models.DateTimeField: datetime_converter }
            field_type = instance._meta.get_field(django_field).__class__
            if val == None:
                return []
            if field_type in converters:
                val = converters[field_type](val)
            else:
                val = unicode(val)
            return [val]
        elif callable(django_field):
            return django_field(instance)
        else:
            return []

    def get_configuration(self):
        """Get the configuration for this indexer, by looking at self.fields.

        """
        fields = {}
        for field in self.fields:
            config = copy.deepcopy(self.defaults)
            (_, search_fieldname, field_config) = self.get_details(field)
            config.update(field_config)
            fields[search_fieldname] = config
        return fields

    def get_current_mapping(self):
        """Get the current mapping for this indexer used by the search engine.

        """
        typename = self.get_typename(self.model)
        return self.client.get_mapping(typename)

    def apply_mapping(self):
        """Apply the configuration for this indexer to the search engine.

        """
        mapping = self.get_configuration()
        typename = self.get_typename(self.model)
        self.client.set_mapping(typename, mapping)

    def make_searcher(self, manager):
        """Make a searcher for the given manager.

        """
        return search.make_searcher(manager, self.model)



"""Utility functions for searchify.

"""

from django.db import models
from django.conf import settings

if hasattr(settings, 'ENABLE_SEARCHIFY') and settings.ENABLE_SEARCHIFY:
    def get_searcher(model_or_instance):
        """Given a model or instance, find the searcher for it.

        """
        if not isinstance(model_or_instance, models.base.ModelBase):
            model = type(model_or_instance)
        else:
            model = model_or_instance
        if not hasattr(model, '_searchify'):
            return None
        if not hasattr(model._searchify, 'searcher'):
            return None
        return model._searchify.searcher

    def get_indexer(model_or_instance):
        """Given a model or instance, find the indexer for it.

        """
        if not isinstance(model_or_instance, models.base.ModelBase):
            model = type(model_or_instance)
        else:
            model = model_or_instance
        if not hasattr(model, '_searchify'):
            return None
        if not hasattr(model._searchify, 'indexer'):
            return None
        return model._searchify.indexer
else:
    get_indexer = get_searcher = lambda x: None


def lookup_model(modeldesc):
    """Convert a packed docid into a Model.

    """
    try:
        (app_label, model_name) = modeldesc.rsplit(".", 1)
    except ValueError:
        return None
    return models.get_model(app_label, model_name)


def get_typename_from_object(instance):
    return '%s.%s' % (
        instance._meta.app_label, instance._meta.object_name,
    )



# search-specific pieces

def make_searcher(manager, model):
    index = get_index(model)
    if not index:
        return None

    client = get_client(index)
    def search(query=None, start=None, end=None, query_filter=None):
        if start==None and end==None:
            # new-style interface. Return an object which responds to slicing and is iterable.
            class SearchResultSet:
                def __init__(self, query, query_filter=None):
                    self.query = query
                    self.query_filter = query_filter
                    self.position = None
                    self.smallest_page = 10
                    # the following are filled out as needed
                    self._results = None # just the current page of results (but contains useful metadata also)
                    self.results = {} # store of index -> decorated Django instance

                def __repr__(self):
                    return "<SearchResultSet()>"

                def __getattr__(self, key):
                    # print "__getattr__(%s)" % key
                    """Provide .-access to aspects of the result. eg: q.doc_count (providing the search provider returns doc_count)."""
                    self._ensure_results()
                    return getattr(self._results, key)
                    
                def _ensure_results(self, start=None, end=None):
                    # print "_ensure_results(%s,%s)" % (start, end)
                    """
                    Call this before any operation to ensure that we've got some useful results available.
                    If called with start=None, then ensure that the result at self.position is available.
                    If not, ensure that the result at [start] or results at [start:end] are available.
                    """
                    if start==None:
                        start=0
                    if end==None:
                        end=start+1
                    # now reduce [start:end] to the single smallest range that we don't already have
                    while self.results.has_key(start) and start < end:
                        start += 1
                    while self.results.has_key(end) and start < end:
                        end -= 1
                    if start==end:
                        # we have everything we need already
                        return
                    if end-start < self.smallest_page:
                        end = start + self.smallest_page
                    
                    self._results = client.search(self.query, self.query_filter, start, end)
                    result_ids = []
                    match_details = {}
                    for item in self._results.results:
                        search_id = item.docid
                        database_name, model_key, django_id = search_id.split('.')
                        if model_key != model.__name__:
                            # FIXME: not right!
                            # Either we need to get this right, or we could filter the query through a boolean to restrict to
                            # this model in the first place. The latter would work better, but requires some thought. (In particular,
                            # semi-constructing queries like this unsettles Richard, so there's probably a reason to avoid it.)
                            #continue
                            pass
                        result_ids.append(long(django_id))
                        match_details[django_id] = item
                    bulk = manager.in_bulk(result_ids)

                    for key, obj in bulk.items():
                        # From Flax we get: data (dict of field->data pairs), id, rank
                        # We only really care about rank at this stage, as we've pulled out the object.
                        if hasattr(model.Searchable, 'match_details_attribute'):
                            match_attr = model.Searchable.match_details_attribute
                        else:
                            match_attr = 'match'
                        if match_attr is not None:
                            setattr(obj, match_attr, match_details[str(obj.pk)])
                        self.results[start] = obj
                        start += 1
                
                def __iter__(self):
                    return self
                    
                def next(self):
                    # print "next()"
                    self._ensure_results()
                    ret = self[self.position]
                    self.position += 1
                    return ret
                    
                def __len__(self):
                    # print "__len__()"
                    self._ensure_results()
                    # this is perhaps not the ideal solution, but should work in general
                    return self.matches_upper_bound
                    
                def __getslice__(self, start, end):
                    # print "__getslice__(%s, %s)" % (start, end)
                    self._ensure_results(start, end)
                    ret = []
                    try:
                        for idx in range(start, end):
                            ret.append(self.results[idx])
                    except KeyError:
                        # Slices fail silently
                        pass
                    return ret
                    
                def __getitem__(self, index):
                    # print "__getitem__(%s)" % index
                    self._ensure_results(index)
                    try:
                        return self.results[index]
                    except KeyError:
                        # Indexing fails noisily
                        raise IndexError('list index out of range')
            
            return SearchResultSet(query, query_filter)
        else:
            # old-style interface
            if start==None:
                start=0
            if end==None:
                end=10
        class QueryResult:
            def __init__(self, results):
                self.results = results
                search_ids = [ item.docid for item in results.results ]
                self._result_ids = []
                self._deets = {}
                for item in results.results:
                    search_id = item.docid
                    database_name, model_key, id = search_id.split('.')
                    if model_key != model.__name__:
                        # FIXME: not right!
                        pass
                    self._result_ids.append(long(id))
                    self._deets[id] = item
                self._bulk = manager.in_bulk(self._result_ids)

            # From Flax, we get: matches_lower_bound, matches_upper_bound, more_matches, matches_estimated, matches_human_readable_estimate
            def __getattr__(self, key):
                """Provide .-access to aspects of the result. eg: q.doc_count (providing the search provider returns doc_count)."""
                return getattr(self.results, key)

            def __len__(self):
                return len(self._result_ids)
            
            def __iter__(self):
                """
                Iterate over the results, in the order they were in the result set.
                Return a decorated object, ie the Django model instance with an extra attribute (default 'match') containing match details (you mostly care about .rank, if provided).
                """
                for key in self._result_ids:
                    obj = self._bulk[long(key)]
                    # From Flax we get: data (dict of field->data pairs), id, rank
                    # We only really care about rank at this stage, as we've pulled out the object.
                    if hasattr(model.Searchable, 'match_details_attribute'):
                        match_attr = model.Searchable.match_details_attribute
                    else:
                        match_attr = 'match'
                    if match_attr is not None:
                        setattr(obj, match_attr, self._deets[str(obj.pk)])
                    yield obj
        
        results = client.search(query, query_filter, start, end)
        return QueryResult(results)
    return search



"""Hooks to ensure that the indexer is informed when an indexed instance
changes.

"""

from django.db.models.signals import post_save, pre_delete, post_delete

from index import get_indexer

def connect_signals():
    post_save.connect(index_hook)
    pre_delete.connect(delete_hook)

def index_hook(sender, **kwargs):
    instance = kwargs['instance']
    indexer = get_indexer(instance)
    if indexer:
        indexer.index_instance(instance)

def delete_hook(sender, **kwargs):
    instance = kwargs['instance']
    indexer = get_indexer(instance)
    if indexer:
        indexer.delete(instance)
        post_delete.connect(post_delete_hook(instance))

def post_delete_hook(instance):
    def hook(sender, **kwargs):
        if kwargs['instance'] == instance:
            indexer = get_indexer(instance)
            if indexer:
                indexer.cascade(instance)
            post_delete.disconnect(hook)
    return hook



# Use setuptools if we can
try:
    from setuptools.core import setup
except ImportError:
    from distutils.core import setup

PACKAGE = 'django-searchify'
VERSION = '0.1'

setup(
    name=PACKAGE, version=VERSION,
    description="Search integration for Django with a focus on indexing",
    packages=[ 'searchify' ],
    license='MIT',
    author='James Aylett',
    author_email='james@tartarus.org',
    install_requires=[
        'Django>=1.3',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Framework :: Django',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
    ],
)
